;;; el-pkgutils.el --- functions that may be useful for packaging 
;;; Emacs Lisp modules.

;; Copyright (C) 2002 ALT Linux Team

;; Author: Ivan Zakharyaschev <imz@altlinux.ru>
;; Last change: 24 Feb 2002

;; (This file is not a part of GNU Emacs.)

;; This code is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This code is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; (along with GNU Emacs); see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; The group of fraction-* functions helps to generate file lists 
;; (particularly for RPM) of .el/.elc fractions.
;;
;; Use a command like this in your spec-file's %install section:
;;   emacs -batch --no-site-file \
;;    -l el-pkgutils  -f batch-fraction-save-lists \
;;    '(el-without-elc "/usr/share/emacs/21.1") "main-filelist"' \
;;    '(el-with-elc "/usr/share/emacs/21.1") "el-filelist"'
;;
;; Alternatively, you can use the shell interface to this module
;; which is described in el-pkgutils.sh.

;;; Code:

(defun replace-in-filename(filename pattern replacement)
  "Replace a part of the string FILENAME that matches PATTERN
   with REPLACEMENT (only once). 

   Case is fixed for the arguments, it is not changed when operating.
   The replacement is inserted literally, without any expansions.

   This function is implemented throug 
   `string-match' and `replace-match' functions. It is an auxilary 
   function required for fraction-* functions from `el-pkgutils' to work."
  (let ((case-fold-search nil))
    (if (string-match pattern filename)
	(replace-match replacement "fixed case" "literal" filename)
      filename)))

(defvar fraction-suffices (list "" ".gz" ".bz2")
  "A list of all suffices that may be added to a filename with an Emacs Lisp 
   module so that it remains an Emacs Lisp module (for example, 
   empty suffix or suffices added to indicate compression).

   This list is used by fraction-* function from `el-pkgutils' to 
   distinguish Emacs Lisp modules. The must contain at least an empty 
   string for the functions to work reasonably.")

(require 'find-lisp)

(defun fraction-list-dirs(directory)
  (let 
      ((result 
	(find-lisp-find-files-internal 
	 directory 
	 'find-lisp-file-predicate-is-directory 
	 'find-lisp-default-directory-predicate)))
  (cons 
   directory
   (delq ; to be sure that directory is listed exactly one time
    directory 
    result))))

(defun fraction-list-files(directory)
  (find-lisp-find-files directory ""))

(defun fraction-build-suffix-regexp(main-suffix)
  (let ((suffices fraction-suffices))
    (concat 
     main-suffix
     "\\("(mapconcat 'regexp-quote suffices "\\|")"\\)"
     "$")))

(defun fraction-list-suffixed-with-test(directory main-suffix 
						  &optional test)
  "The basic function for listing some fractions of files. It distinguishes 
them by matching the tail of a filename against the regexp MAIN-SUFFIX 
and by testing a condition implemented by the function TEST.
(TEST filename) should eval either to an element of the resulting list 
of files (if a condition is met; in the simplest case, just the filename), or nil, which will be discarded.
The search is done inside DIRECTORY (with all subdirectories), and
the value of `fraction-suffices' is used when matching the filenames so that files with those suffics appended count as if they were without them."
    (delq nil 
	  (mapcar (or (bound-and-true-p test) 'identity)
		  (find-lisp-find-files 
		   directory
		   (fraction-build-suffix-regexp main-suffix)))))

(defun fraction-list-el(directory)
  "List all .el files under DIRECTORY."
  (fraction-list-suffixed-with-test directory ".\\.el"))
(defun fraction-list-elc(directory)
  "List all .elc files under DIRECTORY."
  (fraction-list-suffixed-with-test directory ".\\.elc"))
(defun fraction-list-etc(directory)
  "List all non-.el/.elc files under DIRECTORY."
  (let ((el-elc-regexp (fraction-build-suffix-regexp ".\\.elc?")))
    (fraction-list-suffixed-with-test 
     directory ""
     (lambda(f) (if (not (string-match el-elc-regexp f)) f)))))


(defun fraction-has-elc(file)
  "Auxilary function: test whether there is a corresponding
   byte-compiled module along with FILE (which must be .el).

   This function can be used as a parameter to 
   `fraction-list-suffixed-with-test'. It uses the value of 
   `fraction-suffices' to make decisions (test the existence of the
   corresponding .elc file with all the suffices appended."
  (let 
      ((suffices fraction-suffices)
       (file-to-test (replace-in-filename file "\\.el.*$" ".elc")))
    (eval (cons 'or 
		(mapcar (lambda(suffix) 
			  (file-exists-p (concat file-to-test suffix)))
			suffices)))))

(defun fraction-list-el-with-elc(directory)
  "List all .el files under DIRECTORY that have a byte-compiled companion."
  (fraction-list-suffixed-with-test 
   directory 
   ".\\.el"
   (lambda(file) (if (fraction-has-elc file) file))))

(defun fraction-list-el-without-elc(directory)
  "List all .el files under DIRECTORY that don't have a byte-compiled companion."
  (fraction-list-suffixed-with-test 
   directory 
   ".\\.el"
   (lambda(file) (if (not (fraction-has-elc file)) file))))


(defun fraction-list-for-rpm-1(what where)
  (let* 
      ((prefix (expand-file-name (or (getenv "RPM_BUILD_ROOT") "") "/"))
       (prefix-re (concat "^" (regexp-quote prefix))))
    (mapcar
     (lambda(file)
       (replace-in-filename file prefix-re ""))
     (funcall 
      (symbol-function (intern (concat "fraction-list-" (symbol-name what))))
      (concat prefix where)))))
(defun fraction-list-for-rpm(what where &optional set-modifiers)
  "This function should be used from an RPM build environment.
   It lists the files in fraction identified by WHAT found in WHERE.

   WHERE is an absolute directory path (with a leading /); it will be
   prefix with RPM_BUILD_ROOT.
  
   WHAT should be the tail of a corresponding Lisp function name, e.g.
   'el, 'elc, 'el-with-elc etc.

  SET-MODIFIERS is used to prefix entries in the resulting list with some
  modifiers understood by RPM spec-file's %files section. 
  SET-MODIFIERS should be either a string (then it will be used 
  as a prefix for all entries) or a function (SET-MODIFIERS(filename) 
  should eval to the prefix for that file),"
  (let ((set-modifiers-f
	 (cond 
	  ((stringp set-modifiers) (lambda(filename) set-modifiers))
	  ((functionp set-modifiers) set-modifiers)
	  (t (lambda(filename) nil)))))
    (mapcar 
     (lambda(filename) 
       (let ((mods (funcall set-modifiers-f filename)))
	 (if (or (null mods) (zerop (length mods)))
	     filename
	   (concat  mods " " filename))))
     (fraction-list-for-rpm-1 what where))))

(defun fraction-save-list-for-rpm(task)
  "This function should be used from an RPM build environment.
   It saves a list of files of a fraction by appending it to files. These
   files can be used in RPM spec-file's %files section afterwards.

   TASK is a list of the form ( (WHAT WHERE [ MOD ] ) FILE1 FILE2 ...), 
   FILEs are the files for saving the list."
  (let ((result (mapconcat 
		 (lambda(entry) (concat entry "\n"))
		 (apply 'fraction-list-for-rpm (car task))
		 "")))
    (mapc (lambda(place-to-save)
	    (append-to-file result nil place-to-save))
	  (cdr task))))

(defun fraction-save-lists-for-rpm(tasks) (mapc 'fraction-save-list-for-rpm tasks))

(defun batch-fraction-save-lists() 
  "The function to be used from RPM to generate file lists 
of .el/.elc fractions.

Use a command like this in your spec-file's %install section:
   emacs -batch --no-site-file \\
    -l el-pkgutils  -f batch-fraction-save-lists \\
    '(el-without-elc \"/usr/share/emacs/21.1\") \"main-filelist\"' \\
    '(el-with-elc \"/usr/share/emacs/21.1\") \"el-filelist\"' \\
  '(dirs \"/usr/share/emacs/21.1\" \"%dir\") \"main-filelist\" \"el-filelist\"'"
  (mapc (lambda(task-str)
	  (fraction-save-list-for-rpm 
	   (car (read-from-string (concat "(" task-str ")")))))
	command-line-args-left)
  (setq command-line-args-left (list)))

(defun fraction-list-el-sorted-by-name(directory)
  (sort (fraction-list-el directory) 'string<))
(defun fraction-list-elc-sorted-by-name(directory)
  (sort (fraction-list-elc directory) 'string<))

(defun file-size-less-p(f1 f2) 
  (< 
   (nth 7 (file-attributes f1))
   (nth 7 (file-attributes f2))))
(defun fraction-list-el-sorted-by-size(directory)
  (sort (fraction-list-el directory) 'file-size-less-p))
(defun fraction-list-elc-sorted-by-size(directory)
  (sort (fraction-list-elc directory) 'file-size-less-p))
(defun fraction-list-el-without-elc-sorted-by-size(directory)
  (sort (fraction-list-el-without-elc directory) 'file-size-less-p))

(defun append-size-to-filename(file)
  (let 
      ((size-str (number-to-string (nth 7 (file-attributes file)))))
    (concat 
     file 
     (make-string 
      (max 1 (- 70 (length size-str) (length file)))
      (string-to-char " "))
     size-str)))
(defun fraction-list-elc-sorted-by-size*(directory)
  (mapcar 'append-size-to-filename (fraction-list-elc-sorted-by-size directory)))
(defun fraction-list-el-without-elc-sorted-by-size*(directory)
  (mapcar 'append-size-to-filename (fraction-list-el-without-elc-sorted-by-size directory)))

(provide 'el-pkgutils)

; el-pkgutils.el ends here

; Tests:
(quote

(fraction-list-el "/usr/share/emacs/21.1")
(fraction-list-for-rpm 'el-without-elc "/usr/share/emacs/21.1")

(let ((find-lisp-file-predicate )))
(cons)

)
